@@ -12,7 +12,7 @@
 #include <iomanip>
 #include <fstream>
 
-#ifdef HAVE_GL
+#ifdef HAVE_LIBGLM
 #include <glm/glm.hpp>
 #include <glm/gtc/matrix_transform.hpp>
 #include <glm/gtc/type_ptr.hpp>
@@ -20,28 +20,18 @@
 
 using namespace prc;
 #include "material.h"
+
 namespace camp {
 
-#ifdef HAVE_GL
-mem::vector<Material> drawElement::material;
-MaterialMap drawElement::materialMap;
-size_t drawElement::materialIndex;
-
-using gl::modelView;
-//double* Tx=modelView.T;   // x-component of current transform
-//double* Ty=modelView.T+4; // y-component of current transform
-double* Tz=modelView.T+8; // z-component of current transform
-
-#endif
-
+mem::vector<triple> drawElement::centers;
+size_t drawElement::centerIndex=0;
+centerMap drawElement::centermap;
 const triple drawElement::zero;
 
 using vm::array;
+using settings::getSetting;
 
-#ifdef HAVE_GL
-BezierCurve drawSurface::C;
-BezierPatch drawBezierPatch::S;
-BezierTriangle drawBezierTriangle::S;
+#ifdef HAVE_LIBGLM
 
 void storecolor(GLfloat *colors, int i, const vm::array &pens, int j)
 {
@@ -61,55 +51,35 @@ void storecolor(GLfloat *colors, int i, const RGBAColour& p)
   colors[i+3]=p.A;
 }
 
-void clearMaterialBuffer(bool draw)
+void setcolors(const RGBAColour& diffuse, const RGBAColour& emissive,
+               const RGBAColour& specular, double shininess,
+               double metallic, double fresnel0, abs3Doutfile *out)
 {
-  if(draw) drawBezierPatch::S.draw();
-  drawElement::material.clear();
-  drawElement::material.reserve(nmaterials);
-  drawElement::materialMap.clear();
-  drawElement::materialIndex=0;
-}
+  Material m=Material(glm::vec4(diffuse.R,diffuse.G,diffuse.B,diffuse.A),
+                      glm::vec4(emissive.R,emissive.G,emissive.B,emissive.A),
+                      glm::vec4(specular.R,specular.G,specular.B,specular.A),
+                      shininess,metallic,fresnel0);
 
-void setcolors(bool colors,
-               const RGBAColour& diffuse,
-               const RGBAColour& ambient,
-               const RGBAColour& emissive,
-               const RGBAColour& specular, double shininess) 
-{
-  Material m;
-  if(colors) {
-    static glm::vec4 Black(0.0,0.0,0.0,diffuse.A);
-    m=Material(Black,Black,Black,
-               glm::vec4(specular.R,specular.G,specular.B,specular.A),
-               shininess);
-  }  else
-    m=Material(glm::vec4(diffuse.R,diffuse.G,diffuse.B,diffuse.A),
-               glm::vec4(ambient.R,ambient.G,ambient.B,ambient.A),
-               glm::vec4(emissive.R,emissive.G,emissive.B,emissive.A),
-               glm::vec4(specular.R,specular.G,specular.B,specular.A),
-               shininess);
-          
-  MaterialMap::iterator p=drawElement::materialMap.find(m);
-  if(p != drawElement::materialMap.end())
-    drawElement::materialIndex=p->second;
+  auto p=materialMap.find(m);
+  if(p != materialMap.end()) materialIndex=p->second;
   else {
-    drawElement::materialIndex=drawElement::material.size();
-    if(drawElement::materialIndex >= nmaterials)
+    materialIndex=materials.size();
+    if(materialIndex >= nmaterials)
       nmaterials=min(Maxmaterials,2*nmaterials);
-    if(drawElement::materialIndex >= Maxmaterials)
-      clearMaterialBuffer(true);
-    drawElement::material.push_back(m);
-    drawElement::materialMap[m]=drawElement::materialIndex;
+    materials.push_back(m);
+    materialMap[m]=materialIndex;
+    if(out)
+      out->addMaterial(m);
   }
 }
 
-#endif  
+#endif
 
 void drawBezierPatch::bounds(const double* t, bbox3& b)
 {
   double x,y,z;
   double X,Y,Z;
-  
+
   if(straight) {
     triple Vertices[4];
     if(t == NULL) {
@@ -123,48 +93,48 @@ void drawBezierPatch::bounds(const double* t, bbox3& b)
       Vertices[2]=t*controls[12];
       Vertices[3]=t*controls[15];
     }
-    
+
     boundstriples(x,y,z,X,Y,Z,4,Vertices);
   } else {
     double cx[16];
     double cy[16];
     double cz[16];
-  
+
     if(t == NULL) {
-      for(int i=0; i < 16; ++i) {
+      for(unsigned int i=0; i < 16; ++i) {
         triple v=controls[i];
         cx[i]=v.getx();
         cy[i]=v.gety();
         cz[i]=v.getz();
       }
     } else {
-      for(int i=0; i < 16; ++i) {
+      for(unsigned int i=0; i < 16; ++i) {
         triple v=t*controls[i];
         cx[i]=v.getx();
         cy[i]=v.gety();
         cz[i]=v.getz();
       }
     }
-    
+
     double c0=cx[0];
     double fuzz=Fuzz*run::norm(cx,16);
-    x=bound(cx,min,b.empty ? c0 : min(c0,b.left),fuzz,maxdepth);
-    X=bound(cx,max,b.empty ? c0 : max(c0,b.right),fuzz,maxdepth);
-    
+    x=bound(cx,min,c0,fuzz,maxdepth);
+    X=bound(cx,max,c0,fuzz,maxdepth);
+
     c0=cy[0];
     fuzz=Fuzz*run::norm(cy,16);
-    y=bound(cy,min,b.empty ? c0 : min(c0,b.bottom),fuzz,maxdepth);
-    Y=boundtri(cy,max,b.empty ? c0 : max(c0,b.top),fuzz,maxdepth);
-    
+    y=bound(cy,min,c0,fuzz,maxdepth);
+    Y=bound(cy,max,c0,fuzz,maxdepth);
+
     c0=cz[0];
     fuzz=Fuzz*run::norm(cz,16);
-    z=bound(cz,min,b.empty ? c0 : min(c0,b.lower),fuzz,maxdepth);
-    Z=bound(cz,max,b.empty ? c0 : max(c0,b.upper),fuzz,maxdepth);
-  }  
-  
+    z=bound(cz,min,c0,fuzz,maxdepth);
+    Z=bound(cz,max,c0,fuzz,maxdepth);
+  }
+
   b.add(x,y,z);
   b.add(X,Y,Z);
-  
+
   if(t == NULL) {
     Min=triple(x,y,z);
     Max=triple(X,Y,Z);
@@ -172,7 +142,7 @@ void drawBezierPatch::bounds(const double* t, bbox3& b)
 }
 
 void drawBezierPatch::ratio(const double* t, pair &b, double (*m)(double, double),
-                        double fuzz, bool &first)
+                            double fuzz, bool &first)
 {
   triple buf[16];
   triple* Controls;
@@ -185,7 +155,7 @@ void drawBezierPatch::ratio(const double* t, pair &b, double (*m)(double, double
       Controls[12]=t*controls[12];
       Controls[15]=t*controls[15];
     }
-  
+
     triple v=Controls[0];
     double x=xratio(v);
     double y=yratio(v);
@@ -219,7 +189,7 @@ void drawBezierPatch::ratio(const double* t, pair &b, double (*m)(double, double
       b=pair(xratio(v),yratio(v));
       first=false;
     }
-  
+
     b=pair(bound(Controls,m,xratio,b.getx(),fuzz,maxdepth),
            bound(Controls,m,yratio,b.gety(),fuzz,maxdepth));
   }
@@ -227,95 +197,137 @@ void drawBezierPatch::ratio(const double* t, pair &b, double (*m)(double, double
 
 bool drawBezierPatch::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
-  if(invisible || !prc)
+  if(invisible || primitive)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,PRCshininess);
+  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
 
   if(straight) {
     triple vertices[]={controls[0],controls[12],controls[3],controls[15]};
-    if(colors)
-      out->addQuad(vertices,colors);
-    else
+    if(colors) {
+      prc::RGBAColour Colors[]={colors[0],colors[1],colors[3],colors[2]};
+      out->addQuad(vertices,Colors);
+    } else
       out->addRectangle(vertices,m);
   } else
     out->addPatch(controls,m);
-                    
+
+  return true;
+}
+
+bool drawBezierPatch::write(abs3Doutfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible || primitive)
+    return true;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  if(billboard) {
+    meshinit();
+    drawElement::centerIndex=centerIndex;
+  } else drawElement::centerIndex=0;
+
+  out->precision(digits);
+  if(straight) {
+    triple Controls[]={controls[0],controls[12],controls[15],controls[3]};
+    out->addStraightPatch(Controls,colors);
+  } else {
+    double prerender=renderResolution();
+    if(prerender) {
+      GLfloat c[16];
+      if(colors)
+        for(size_t i=0; i < 4; ++i)
+          storecolor(c,4*i,colors[i]);
+      S.init(prerender,colors ? c : NULL);
+      S.render(controls,straight,c);
+      drawTriangles dt(S.data,center,colors,diffuse,emissive,specular,opacity,
+                       shininess,metallic,fresnel0,interaction,invisible,
+                       Min,Max);
+      dt.write(out);
+    } else
+      out->addPatch(controls,colors);
+  }
+  out->precision(getSetting<Int>("digits"));
+
+#endif
   return true;
 }
 
 void drawBezierPatch::render(double size2, const triple& b, const triple& B,
-                             double perspective, bool transparent)
+                             double perspective, bool remesh)
 {
 #ifdef HAVE_GL
-  if(invisible || 
-     ((colors ? colors[0].A+colors[1].A+colors[2].A+colors[3].A < 4.0 :
-       diffuse.A < 1.0) ^ transparent)) return;
-  
-  const bool billboard=interaction == BILLBOARD &&
-    !settings::getSetting<bool>("offscreen");
-  triple m,M;
-  
-  double f,F,s;
-  if(perspective) {
-    f=Min.getz()*perspective;
-    F=Max.getz()*perspective;
-    m=triple(min(f*b.getx(),F*b.getx()),min(f*b.gety(),F*b.gety()),b.getz());
-    M=triple(max(f*B.getx(),F*B.getx()),max(f*B.gety(),F*B.gety()),B.getz());
-    s=max(f,F);
-  } else {
-    m=b;
-    M=B;
-    s=1.0;
+  if(invisible) return;
+  transparent=colors ? colors[0].A+colors[1].A+colors[2].A+colors[3].A < 4.0 :
+    diffuse.A < 1.0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0);
+
+  if(transparent)
+    setMaterial(transparentData,drawTransparent);
+  else {
+    if(colors)
+      setMaterial(colorData,drawColor);
+    else
+      setMaterial(materialData,drawMaterial);
   }
-  
-  const pair size3(s*(B.getx()-b.getx()),s*(B.gety()-b.gety()));
 
-  bbox3 box(m,M);
-  box.transform(modelView.Tinv);
-  m=box.Min();
-  M=box.Max();
-  
-  if(!billboard && (Max.getx() < m.getx() || Min.getx() > M.getx() ||
-                    Max.gety() < m.gety() || Min.gety() > M.gety() ||
-                    Max.getz() < m.getz() || Min.getz() > M.getz()))
+  bool offscreen;
+  if(billboard) {
+    drawElement::centerIndex=centerIndex;
+    BB.init(center);
+    offscreen=bbox2(Min,Max,BB).offscreen();
+  } else
+    offscreen=bbox2(Min,Max).offscreen();
+
+  if(offscreen) { // Fully offscreen
+    S.Onscreen=false;
+    S.data.clear();
+    S.transparent=transparent;
+    S.color=colors;
+    S.notRendered();
     return;
+  }
 
-  setcolors(colors,diffuse,ambient,emissive,specular,shininess);
-  
-  if(billboard) BB.init(center);
-  
-  GLfloat c[16];
-  if(colors)
-    for(size_t i=0; i < 4; ++i)
-      storecolor(c,4*i,colors[i]);
-  
   triple *Controls;
   triple Controls0[16];
   if(billboard) {
     Controls=Controls0;
-    for(size_t i=0; i < 16; i++)
+    for(size_t i=0; i < 16; i++) {
       Controls[i]=BB.transform(controls[i]);
-  } else
-    Controls=controls;
-    
-  if(gl::outlinemode) {
-    triple edge0[]={Controls[0],Controls[4],Controls[8],Controls[12]};
-    C.queue(edge0,straight,size3.length()/size2,m,M);
-    triple edge1[]={Controls[12],Controls[13],Controls[14],Controls[15]};
-    C.queue(edge1,straight,size3.length()/size2,m,M);
-    triple edge2[]={Controls[15],Controls[11],Controls[7],Controls[3]};
-    C.queue(edge2,straight,size3.length()/size2,m,M);
-    triple edge3[]={Controls[3],Controls[2],Controls[1],Controls[0]};
-    C.queue(edge3,straight,size3.length()/size2,m,M);
-    C.draw();
+    }
   } else {
-    S.queue(Controls,straight,size3.length()/size2,m,M,transparent,
+    Controls=controls;
+    if(!remesh && S.Onscreen) { // Fully onscreen; no need to re-render
+      S.append();
+      return;
+    }
+  }
+
+  double s=perspective ? Min.getz()*perspective : 1.0; // Move to glrender
+
+  const pair size3(s*(B.getx()-b.getx()),s*(B.gety()-b.gety()));
+
+  if(gl::outlinemode) {
+    setMaterial(material1Data,drawMaterial);
+    triple edge0[]={Controls[0],Controls[4],Controls[8],Controls[12]};
+    C.queue(edge0,straight,size3.length()/size2);
+    triple edge1[]={Controls[12],Controls[13],Controls[14],Controls[15]};
+    C.queue(edge1,straight,size3.length()/size2);
+    triple edge2[]={Controls[15],Controls[11],Controls[7],Controls[3]};
+    C.queue(edge2,straight,size3.length()/size2);
+    triple edge3[]={Controls[3],Controls[2],Controls[1],Controls[0]};
+    C.queue(edge3,straight,size3.length()/size2);
+  } else {
+    GLfloat c[16];
+    if(colors)
+      for(size_t i=0; i < 4; ++i)
+        storecolor(c,4*i,colors[i]);
+
+    S.queue(Controls,straight,size3.length()/size2,transparent,
             colors ? c : NULL);
-    if(BezierPatch::nvertices >= gl::maxvertices)
-      drawBezierPatch::S.drawMaterials();
-    if(BezierPatch::Nvertices >= gl::maxvertices)
-      drawBezierPatch::S.drawColors();
   }
 #endif
 }
@@ -324,7 +336,7 @@ drawElement *drawBezierPatch::transformed(const double* t)
 {
   return new drawBezierPatch(t,this);
 }
-  
+
 void drawBezierTriangle::bounds(const double* t, bbox3& b)
 {
   double x,y,z;
@@ -341,13 +353,13 @@ void drawBezierTriangle::bounds(const double* t, bbox3& b)
       Vertices[1]=t*controls[6];
       Vertices[2]=t*controls[9];
     }
-  
+
     boundstriples(x,y,z,X,Y,Z,3,Vertices);
-  } else {  
+  } else {
     double cx[10];
     double cy[10];
     double cz[10];
-    
+
     if(t == NULL) {
       for(unsigned int i=0; i < 10; ++i) {
         triple v=controls[i];
@@ -363,23 +375,23 @@ void drawBezierTriangle::bounds(const double* t, bbox3& b)
         cz[i]=v.getz();
       }
     }
-    
+
     double c0=cx[0];
     double fuzz=Fuzz*run::norm(cx,10);
-    x=boundtri(cx,min,b.empty ? c0 : min(c0,b.left),fuzz,maxdepth);
-    X=boundtri(cx,max,b.empty ? c0 : max(c0,b.right),fuzz,maxdepth);
-    
+    x=boundtri(cx,min,c0,fuzz,maxdepth);
+    X=boundtri(cx,max,c0,fuzz,maxdepth);
+
     c0=cy[0];
     fuzz=Fuzz*run::norm(cy,10);
-    y=boundtri(cy,min,b.empty ? c0 : min(c0,b.bottom),fuzz,maxdepth);
-    Y=boundtri(cy,max,b.empty ? c0 : max(c0,b.top),fuzz,maxdepth);
-    
+    y=boundtri(cy,min,c0,fuzz,maxdepth);
+    Y=boundtri(cy,max,c0,fuzz,maxdepth);
+
     c0=cz[0];
     fuzz=Fuzz*run::norm(cz,10);
-    z=boundtri(cz,min,b.empty ? c0 : min(c0,b.lower),fuzz,maxdepth);
-    Z=boundtri(cz,max,b.empty ? c0 : max(c0,b.upper),fuzz,maxdepth);
+    z=boundtri(cz,min,c0,fuzz,maxdepth);
+    Z=boundtri(cz,max,c0,fuzz,maxdepth);
   }
-    
+
   b.add(x,y,z);
   b.add(X,Y,Z);
 
@@ -403,7 +415,7 @@ void drawBezierTriangle::ratio(const double* t, pair &b,
       Controls[6]=t*controls[6];
       Controls[9]=t*controls[9];
     }
-  
+
     triple v=Controls[0];
     double x=xratio(v);
     double y=yratio(v);
@@ -434,20 +446,21 @@ void drawBezierTriangle::ratio(const double* t, pair &b,
       b=pair(xratio(v),yratio(v));
       first=false;
     }
-  
+
     b=pair(boundtri(Controls,m,xratio,b.getx(),fuzz,maxdepth),
            boundtri(Controls,m,yratio,b.gety(),fuzz,maxdepth));
   }
 }
 
-bool drawBezierTriangle::write(prcfile *out, unsigned int *, double, 
+bool drawBezierTriangle::write(prcfile *out, unsigned int *, double,
                                groupsmap&)
 {
-  if(invisible)
+  if(invisible || primitive)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,PRCshininess);
-  
+  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
+
   static const double third=1.0/3.0;
   static const double third2=2.0/3.0;
   triple Controls[]={controls[0],controls[0],controls[0],controls[0],
@@ -459,76 +472,121 @@ bool drawBezierTriangle::write(prcfile *out, unsigned int *, double,
                      controls[5],controls[6],controls[7],
                      controls[8],controls[9]};
   out->addPatch(Controls,m);
-                    
+
+  return true;
+}
+
+bool drawBezierTriangle::write(abs3Doutfile *out)
+{
+#ifdef HAVE_LIBGLM
+
+  if(invisible || primitive)
+    return true;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  if(billboard) {
+    meshinit();
+    drawElement::centerIndex=centerIndex;
+  } else drawElement::centerIndex=0;
+
+  out->precision(digits);
+  if(straight) {
+    triple Controls[]={controls[0],controls[6],controls[9]};
+    out->addStraightBezierTriangle(Controls,colors);
+  } else {
+    double prerender=renderResolution();
+    if(prerender) {
+      GLfloat c[16];
+      if(colors)
+        for(size_t i=0; i < 4; ++i)
+          storecolor(c,4*i,colors[i]);
+      S.init(prerender,colors ? c : NULL);
+      S.render(controls,straight,c);
+      drawTriangles dt(S.data,center,colors,diffuse,emissive,specular,opacity,
+                       shininess,metallic,fresnel0,interaction,invisible,
+                       Min,Max);
+      dt.write(out);
+    } else
+      out->addBezierTriangle(controls,colors);
+  }
+  out->precision(getSetting<Int>("digits"));
+
+#endif
   return true;
 }
 
 void drawBezierTriangle::render(double size2, const triple& b, const triple& B,
-                                double perspective, bool transparent)
+                                double perspective, bool remesh)
 {
 #ifdef HAVE_GL
-  if(invisible || 
-     ((colors ? colors[0].A+colors[1].A+colors[2].A < 3.0 :
-       diffuse.A < 1.0) ^ transparent)) return;
-  
-  const bool billboard=interaction == BILLBOARD &&
-    !settings::getSetting<bool>("offscreen");
-  triple m,M;
-  
-  double f,F,s;
-  if(perspective) {
-    f=Min.getz()*perspective;
-    F=Max.getz()*perspective;
-    m=triple(min(f*b.getx(),F*b.getx()),min(f*b.gety(),F*b.gety()),b.getz());
-    M=triple(max(f*B.getx(),F*B.getx()),max(f*B.gety(),F*B.gety()),B.getz());
-    s=max(f,F);
-  } else {
-    m=b;
-    M=B;
-    s=1.0;
+  if(invisible) return;
+  transparent=colors ? colors[0].A+colors[1].A+colors[2].A < 3.0 :
+    diffuse.A < 1.0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0);
+
+  if(transparent)
+    setMaterial(transparentData,drawTransparent);
+  else {
+    if(colors)
+      setMaterial(colorData,drawColor);
+    else
+      setMaterial(materialData,drawMaterial);
   }
-  
-  const pair size3(s*(B.getx()-b.getx()),s*(B.gety()-b.gety()));
 
-  bbox3 box(m,M);
-  box.transform(modelView.Tinv);
-  m=box.Min();
-  M=box.Max();
+  bool offscreen;
+  if(billboard) {
+    drawElement::centerIndex=centerIndex;
+    BB.init(center);
+    offscreen=bbox2(Min,Max,BB).offscreen();
+  } else
+    offscreen=bbox2(Min,Max).offscreen();
 
-  if(!billboard && (Max.getx() < m.getx() || Min.getx() > M.getx() ||
-                    Max.gety() < m.gety() || Min.gety() > M.gety() ||
-                    Max.getz() < m.getz() || Min.getz() > M.getz()))
+  if(offscreen) { // Fully offscreen
+    S.Onscreen=false;
+    S.data.clear();
+    S.transparent=transparent;
+    S.color=colors;
+    S.notRendered();
     return;
+  }
 
-  setcolors(colors,diffuse,ambient,emissive,specular,shininess);
-  
-  if(billboard) BB.init(center);
-  
-  GLfloat v[12];
-  if(colors)
-    for(size_t i=0; i < 3; ++i)
-      storecolor(v,4*i,colors[i]);
-    
   triple *Controls;
   triple Controls0[10];
   if(billboard) {
     Controls=Controls0;
     for(size_t i=0; i < 10; i++)
       Controls[i]=BB.transform(controls[i]);
-  } else 
+  } else {
+    if(!remesh && S.Onscreen) { // Fully onscreen; no need to re-render
+      S.append();
+      return;
+    }
     Controls=controls;
-  
+  }
+
+  double s=perspective ? Min.getz()*perspective : 1.0; // Move to glrender
+
+  const pair size3(s*(B.getx()-b.getx()),s*(B.gety()-b.gety()));
+
   if(gl::outlinemode) {
+    setMaterial(material1Data,drawMaterial);
     triple edge0[]={Controls[0],Controls[1],Controls[3],Controls[6]};
-    C.queue(edge0,straight,size3.length()/size2,m,M);
+    C.queue(edge0,straight,size3.length()/size2);
     triple edge1[]={Controls[6],Controls[7],Controls[8],Controls[9]};
-    C.queue(edge1,straight,size3.length()/size2,m,M);
+    C.queue(edge1,straight,size3.length()/size2);
     triple edge2[]={Controls[9],Controls[5],Controls[2],Controls[0]};
-    C.queue(edge2,straight,size3.length()/size2,m,M);
-    C.draw();
-  } else
-    S.queue(Controls,straight,size3.length()/size2,m,M,transparent,
-            colors ? v : NULL);
+    C.queue(edge2,straight,size3.length()/size2);
+  } else {
+    GLfloat c[12];
+    if(colors)
+      for(size_t i=0; i < 3; ++i)
+        storecolor(c,4*i,colors[i]);
+
+    S.queue(Controls,straight,size3.length()/size2,transparent,
+            colors ? c : NULL);
+  }
 #endif
 }
 
@@ -536,15 +594,16 @@ drawElement *drawBezierTriangle::transformed(const double* t)
 {
   return new drawBezierTriangle(t,this);
 }
-  
+
 bool drawNurbs::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,PRCshininess);
+  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
   out->addSurface(udegree,vdegree,nu,nv,controls,uknots,vknots,m,weights);
-  
+
   return true;
 }
 
@@ -553,7 +612,7 @@ void drawNurbs::bounds(const double* t, bbox3& b)
 {
   double x,y,z;
   double X,Y,Z;
-  
+
   const size_t n=nu*nv;
   triple* Controls;
   if(t == NULL) Controls=controls;
@@ -564,10 +623,10 @@ void drawNurbs::bounds(const double* t, bbox3& b)
   }
 
   boundstriples(x,y,z,X,Y,Z,n,Controls);
-  
+
   b.add(x,y,z);
   b.add(X,Y,Z);
-  
+
   if(t == NULL) {
     Min=triple(x,y,z);
     Max=triple(X,Y,Z);
@@ -583,7 +642,7 @@ void drawNurbs::ratio(const double *t, pair &b, double (*m)(double, double),
                       double, bool &first)
 {
   const size_t n=nu*nv;
-  
+
   triple* Controls;
   if(t == NULL) Controls=controls;
   else {
@@ -597,7 +656,7 @@ void drawNurbs::ratio(const double *t, pair &b, double (*m)(double, double),
     triple v=Controls[0];
     b=pair(xratio(v),yratio(v));
   }
-  
+
   double x=b.getx();
   double y=b.gety();
   for(size_t i=0; i < n; ++i) {
@@ -606,7 +665,7 @@ void drawNurbs::ratio(const double *t, pair &b, double (*m)(double, double),
     y=m(y,yratio(v));
   }
   b=pair(x,y);
-  
+
   if(t != NULL)
     delete[] Controls;
 }
@@ -618,58 +677,46 @@ void drawNurbs::displacement()
   size_t n=nu*nv;
   size_t nuknots=udegree+nu+1;
   size_t nvknots=vdegree+nv+1;
-    
+
   if(Controls == NULL) {
     Controls=new(UseGC)  GLfloat[(weights ? 4 : 3)*n];
     uKnots=new(UseGC) GLfloat[nuknots];
     vKnots=new(UseGC) GLfloat[nvknots];
   }
-  
+
   if(weights)
     for(size_t i=0; i < n; ++i)
       store(Controls+4*i,controls[i],weights[i]);
   else
     for(size_t i=0; i < n; ++i)
       store(Controls+3*i,controls[i]);
-  
+
   for(size_t i=0; i < nuknots; ++i)
     uKnots[i]=uknots[i];
   for(size_t i=0; i < nvknots; ++i)
     vKnots[i]=vknots[i];
-#endif  
+#endif
 }
 
-void drawNurbs::render(double size2, const triple& Min, const triple& Max,
-                       double perspective, bool transparent)
+void drawNurbs::render(double size2, const triple& b, const triple& B,
+                       double perspective, bool remesh)
 {
-#ifdef HAVE_GL
-  if(invisible || ((colors ? colors[3]+colors[7]+colors[11]+colors[15] < 4.0
-                    : diffuse.A < 1.0) ^ transparent)) return;
+// TODO: implement NURBS renderer
+}
 
-  bbox3 B(this->Min,this->Max);
-  B.transform(modelView.T);
-    
-  triple m=B.Min();
-  triple M=B.Max();
-  
-  if(perspective) {
-    double f=m.getz()*perspective;
-    double F=M.getz()*perspective;
-    if(M.getx() < min(f*Min.getx(),F*Min.getx()) || 
-       m.getx() > max(f*Max.getx(),F*Max.getx()) ||
-       M.gety() < min(f*Min.gety(),F*Min.gety()) ||
-       m.gety() > max(f*Max.gety(),F*Max.gety()) ||
-       M.getz() < Min.getz() ||
-       m.getz() > Max.getz()) return;
-  } else {
-    if(M.getx() < Min.getx() || m.getx() > Max.getx() ||
-       M.gety() < Min.gety() || m.gety() > Max.gety() ||
-       M.getz() < Min.getz() || m.getz() > Max.getz()) return;
+void drawPRC::P(triple& t, double x, double y, double z)
+{
+  if(T == NULL) {
+    t=triple(x,y,z);
+    return;
   }
 
-  setcolors(colors,diffuse,ambient,emissive,specular,shininess);
-// TODO: implement NURBS renderer
-#endif
+  double f=T[12]*x+T[13]*y+T[14]*z+T[15];
+  if(f == 0.0) run::dividebyzero();
+  f=1.0/f;
+
+  t=triple((T[0]*x+T[1]*y+T[2]*z+T[3])*f,(T[4]*x+T[5]*y+T[6]*z+T[7])*f,
+           (T[8]*x+T[9]*y+T[10]*z+T[11])*f);
 }
 
 void drawSphere::P(triple& t, double x, double y, double z)
@@ -677,18 +724,7 @@ void drawSphere::P(triple& t, double x, double y, double z)
   if(half) {
     double temp=z; z=x; x=-temp;
   }
-  
-  if(T == NULL) {
-    t=triple(x,y,z);
-    return;
-  }
-  
-  double f=T[12]*x+T[13]*y+T[14]*z+T[15];
-  if(f == 0.0) run::dividebyzero();
-  f=1.0/f;
-  
-  t=triple((T[0]*x+T[1]*y+T[2]*z+T[3])*f,(T[4]*x+T[5]*y+T[6]*z+T[7])*f,
-           (T[8]*x+T[9]*y+T[10]*z+T[11])*f);
+  drawPRC::P(t,x,y,z);
 }
 
 bool drawSphere::write(prcfile *out, unsigned int *, double, groupsmap&)
@@ -696,12 +732,13 @@ bool drawSphere::write(prcfile *out, unsigned int *, double, groupsmap&)
   if(invisible)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,shininess);
-  
+  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
+
   switch(type) {
     case 0: // PRCsphere
     {
-      if(half) 
+      if(half)
         out->addHemisphere(1.0,m,NULL,NULL,NULL,1.0,T);
       else
         out->addSphere(1.0,m,NULL,NULL,NULL,1.0,T);
@@ -717,9 +754,9 @@ bool drawSphere::write(prcfile *out, unsigned int *, double, groupsmap&)
 
 // NURBS representation of a sphere using 10 distinct control points
 // K. Qin, J. Comp. Sci. and Tech. 12, 210-216 (1997).
-  
+
       triple N,S,P1,P2,P3,P4,P5,P6,P7,P8;
-  
+
       P(N,0.0,0.0,1.0);
       P(P1,-2.0,-2.0,1.0);
       P(P2,-2.0,-2.0,-1.0);
@@ -730,27 +767,52 @@ bool drawSphere::write(prcfile *out, unsigned int *, double, groupsmap&)
       P(P6,2.0,2.0,-1.0);
       P(P7,-2.0,2.0,1.0);
       P(P8,-2.0,2.0,-1.0);
-        
+
       triple p0[]={N,P1,P2,S,
                    N,P3,P4,S,
                    N,P5,P6,S,
                    N,P7,P8,S,
                    N,P1,P2,S,
                    N,P3,P4,S};
-   
+
       out->addSurface(2,3,3,4,p0,uknot,vknot,m,Weights);
       out->addSurface(2,3,3,4,p0+4,uknot,vknot,m,Weights);
       if(!half) {
         out->addSurface(2,3,3,4,p0+8,uknot,vknot,m,Weights);
         out->addSurface(2,3,3,4,p0+12,uknot,vknot,m,Weights);
       }
-      
+
       break;
     }
     default:
       reportError("Invalid sphere type");
   }
-  
+
+  return true;
+}
+
+bool drawSphere::write(abs3Doutfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  drawElement::centerIndex=0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  triple O,E;
+  P(E,1.0,0.0,0.0);
+  P(O,0.0,0.0,0.0);
+  triple X=E-O;
+  double r=length(X);
+
+  if(half)
+    out->addHemisphere(O,r,X.polar(false),X.azimuth(false));
+  else
+    out->addSphere(O,r);
+
+#endif
   return true;
 }
 
@@ -759,72 +821,99 @@ bool drawCylinder::write(prcfile *out, unsigned int *, double, groupsmap&)
   if(invisible)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,shininess);
-  
+  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
+
   out->addCylinder(1.0,1.0,m,NULL,NULL,NULL,1.0,T);
-  
+
   return true;
 }
-  
+
+bool drawCylinder::write(abs3Doutfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  drawElement::centerIndex=0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  triple E,H,O;
+  P(E,1.0,0.0,0.0);
+  P(H,0.0,0.0,1.0);
+  P(O,0.0,0.0,0.0);
+  triple X=E-O;
+  triple Z=H-O;
+  double r=length(X);
+  double h=length(Z);
+
+  out->addCylinder(O,r,h,Z.polar(false),Z.azimuth(false),core);
+
+#endif
+  return true;
+}
+
 bool drawDisk::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,shininess);
-  
+  RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+  PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
+
   out->addDisk(1.0,m,NULL,NULL,NULL,1.0,T);
-  
+
   return true;
 }
-  
-bool drawTube::write(prcfile *out, unsigned int *, double, groupsmap&)
+
+bool drawDisk::write(abs3Doutfile *out)
 {
+#ifdef HAVE_LIBGLM
   if(invisible)
     return true;
 
-  PRCmaterial m(ambient,diffuse,emissive,specular,opacity,shininess);
-  
-  Int n=center.length();
-  
-  if(center.piecewisestraight()) {
-    triple *centerControls=new(UseGC) triple[n+1];
-    for(Int i=0; i <= n; ++i)
-      centerControls[i]=center.point(i);
-    size_t N=n+1;
-    triple *controls=new(UseGC) triple[N];
-    for(Int i=0; i <= n; ++i)
-      controls[i]=g.point(i);
-    out->addTube(N,centerControls,controls,true,m);
-  } else {
-    size_t N=3*n+1;
-    triple *centerControls=new(UseGC) triple[N];
-    centerControls[0]=center.point((Int) 0);
-    centerControls[1]=center.postcontrol((Int) 0);
-    size_t k=1;
-    for(Int i=1; i < n; ++i) {
-      centerControls[++k]=center.precontrol(i);
-      centerControls[++k]=center.point(i);
-      centerControls[++k]=center.postcontrol(i);
-    }
-    centerControls[++k]=center.precontrol(n);
-    centerControls[++k]=center.point(n);
-    
-    triple *controls=new(UseGC) triple[N];
-    controls[0]=g.point((Int) 0);
-    controls[1]=g.postcontrol((Int) 0);
-    k=1;
-    for(Int i=1; i < n; ++i) {
-      controls[++k]=g.precontrol(i);
-      controls[++k]=g.point(i);
-      controls[++k]=g.postcontrol(i);
-    }
-    controls[++k]=g.precontrol(n);
-    controls[++k]=g.point(n);
-    
-    out->addTube(N,centerControls,controls,false,m);
-  }
-      
+  drawElement::centerIndex=0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  triple E,H,O;
+  P(E,1.0,0.0,0.0);
+  P(H,0.0,0.0,1.0);
+  P(O,0.0,0.0,0.0);
+  triple X=E-O;
+  triple Z=H-O;
+  double r=length(X);
+
+  out->addDisk(O,r,Z.polar(false),Z.azimuth(false));
+
+#endif
+  return true;
+}
+
+bool drawTube::write(abs3Doutfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  drawElement::centerIndex=0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+
+  bbox3 b;
+  b.add(T*m);
+  b.add(T*triple(m.getx(),m.gety(),M.getz()));
+  b.add(T*triple(m.getx(),M.gety(),m.getz()));
+  b.add(T*triple(m.getx(),M.gety(),M.getz()));
+  b.add(T*triple(M.getx(),m.gety(),m.getz()));
+  b.add(T*triple(M.getx(),m.gety(),M.getz()));
+  b.add(T*triple(M.getx(),M.gety(),m.getz()));
+  b.add(T*M);
+
+  out->addTube(g,width,core);
+
+#endif
   return true;
 }
 
@@ -870,7 +959,7 @@ void drawBaseTriangles::ratio(const double* t, pair &b,
   }
 
   ratiotriples(b,m,first,nP,tP);
-  
+
   if(t != NULL)
     delete[] tP;
 }
@@ -879,57 +968,86 @@ bool drawTriangles::write(prcfile *out, unsigned int *, double, groupsmap&)
 {
   if(invisible)
     return true;
-  
+
   if(nC) {
     const RGBAColour white(1,1,1,opacity);
     const RGBAColour black(0,0,0,opacity);
-    const PRCmaterial m(black,white,black,specular,opacity,PRCshininess);
+    const PRCmaterial m(black,white,black,specular,opacity,shininess);
     out->addTriangles(nP,P,nI,PI,m,nN,N,NI,0,NULL,NULL,nC,C,CI,0,NULL,NULL,30);
   } else {
-    const PRCmaterial m(ambient,diffuse,emissive,specular,opacity,PRCshininess);
+    RGBAColour Black(0.0,0.0,0.0,diffuse.A);
+    const PRCmaterial m(Black,diffuse,emissive,specular,opacity,shininess);
     out->addTriangles(nP,P,nI,PI,m,nN,N,NI,0,NULL,NULL,0,NULL,NULL,0,NULL,NULL,30);
   }
 
   return true;
 }
 
-void drawTriangles::render(double size2, const triple& Min,
-                           const triple& Max, double perspective,
-                           bool transparent)
+bool drawTriangles::write(abs3Doutfile *out)
+{
+#ifdef HAVE_LIBGLM
+  if(invisible)
+    return true;
+
+  if(billboard) {
+    meshinit();
+    drawElement::centerIndex=centerIndex;
+  } else drawElement::centerIndex=0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0,out);
+  out->addTriangles(nP,P,nN,N,nC,C,nI,PI,NI,CI);
+#endif
+  return true;
+}
+
+void drawTriangles::render(double size2, const triple& b,
+                           const triple& B, double perspective,
+                           bool remesh)
 {
 #ifdef HAVE_GL
-  if(invisible)
+  if(invisible) return;
+  transparent=diffuse.A < 1.0;
+
+  setcolors(diffuse,emissive,specular,shininess,metallic,fresnel0);
+
+  if(transparent)
+    setMaterial(transparentData,drawTransparent);
+  else
+    setMaterial(triangleData,drawTriangle);
+
+  bool offscreen;
+  if(billboard) {
+    drawElement::centerIndex=centerIndex;
+    BB.init(center);
+    offscreen=bbox2(Min,Max,BB).offscreen();
+  } else
+    offscreen=bbox2(Min,Max).offscreen();
+
+  if(offscreen) { // Fully offscreen
+    R.Onscreen=false;
+    R.data.clear();
+    R.transparent=transparent;
+    R.notRendered();
     return;
-
-  if(invisible || ((diffuse.A < 1.0) ^ transparent)) return;
-
-  triple m,M;
-
-  bbox3 B(this->Min,this->Max);
-  B.transform(modelView.T);
-
-  m=B.Min();
-  M=B.Max();
-
-  if(perspective) {
-    const double f=m.getz()*perspective;
-    const double F=M.getz()*perspective;
-    if((M.getx() < min(f*Min.getx(),F*Min.getx()) ||
-        m.getx() > max(f*Max.getx(),F*Max.getx()) ||
-        M.gety() < min(f*Min.gety(),F*Min.gety()) ||
-        m.gety() > max(f*Max.gety(),F*Max.gety()) ||
-        M.getz() < Min.getz() ||
-        m.getz() > Max.getz()))
-      return;
-  } else {
-    if((M.getx() < Min.getx() || m.getx() > Max.getx() ||
-        M.gety() < Min.gety() || m.gety() > Max.gety() ||
-        M.getz() < Min.getz() || m.getz() > Max.getz()))
-      return;
   }
 
-  setcolors(nC,diffuse,ambient,emissive,specular,shininess);
-  R.queue(nP,P,nN,N,nC,C,nI,PI,NI,CI,transparent);
+  triple *P0;
+  if(billboard) {
+    P0=new triple [nP];
+    for(size_t i=0; i < nP; i++)
+      P0[i]=BB.transform(P[i]);
+  } else {
+    if(!remesh && R.Onscreen) { // Fully onscreen; no need to re-render
+      R.append();
+      return;
+    }
+    P0=P;
+  }
+
+  R.queue(nP,P0,nN,N,nC,C,nI,PI,NI,CI,transparent);
+
+  if(billboard)
+    delete [] P0;
 #endif
 }
 
