@@ -8,6 +8,7 @@
 #include <fstream>
 #include <cstring>
 #include <sys/time.h>
+#include <unistd.h>
 
 #include "common.h"
 #include "locate.h"
@@ -25,6 +26,7 @@
 extern uint32_t CLZ(uint32_t a);
 
 bool GPUindexing;
+bool GPUcompress;
 
 namespace gl {
 #ifdef HAVE_PTHREAD
@@ -63,7 +65,7 @@ pthread_t mainthread;
 #endif
 
 using settings::locateFile;
-using utils::seconds;
+using utils::stopWatch;
 
 #endif // HAVE_GL
 
@@ -80,19 +82,24 @@ GLint countShader;
 GLint transparentShader;
 GLint blendShader;
 GLint zeroShader;
+GLint compressShader;
 GLint sum1Shader;
 GLint sum2Shader;
+GLint sum2fastShader;
 GLint sum3Shader;
 
+GLuint fragments;
+
 GLuint offsetBuffer;
+GLuint indexBuffer;
+GLuint elementsBuffer;
 GLuint countBuffer;
-GLuint maxBuffer;
-GLuint localSumBuffer;
 GLuint globalSumBuffer;
 GLuint fragmentBuffer;
 GLuint depthBuffer;
 GLuint opaqueBuffer;
 GLuint opaqueDepthBuffer;
+GLuint feedbackBuffer;
 
 bool ssbo;
 bool interlock;
@@ -160,8 +167,10 @@ GLint gs2;
 GLint gs;
 GLint g;
 GLuint processors;
-GLuint localsize;
-GLint maxgroups;
+GLuint localSize;
+GLuint blockSize;
+GLuint groupSize;
+//GLint maxgroups;
 GLuint maxSize;
 
 bool outlinemode=false;
@@ -204,6 +213,7 @@ string Format;
 int fullWidth,fullHeight;
 int Width,Height;
 GLuint pixels;
+GLuint elements;
 GLuint lastpixels;
 double oWidth,oHeight;
 int screenWidth,screenHeight;
@@ -422,7 +432,6 @@ double T[16];
 
 #ifdef HAVE_LIBGLUT
 timeval lasttime;
-timeval lastframetime;
 int oldWidth,oldHeight;
 
 bool queueScreen=false;
@@ -529,40 +538,42 @@ void noShaders()
   exit(-1);
 }
 
-// Return ceil(log2(n)) where n is a 32 bit unsigned integer.
-uint32_t ceillog2(uint32_t n)
-{
-  return 32-CLZ(n-1);
-}
-
 void initComputeShaders()
 {
   string sum1=locateFile("shaders/sum1.glsl");
   string sum2=locateFile("shaders/sum2.glsl");
+  string sum2fast=locateFile("shaders/sum2fast.glsl");
   string sum3=locateFile("shaders/sum3.glsl");
 
-  if(sum1.empty() || sum2.empty() || sum3.empty())
+  if(sum1.empty() || sum2.empty() || sum2fast.empty() || sum3.empty())
     noShaders();
 
   std::vector<ShaderfileModePair> shaders(1);
   std::vector<std::string> shaderParams;
 
   shaders[0]=ShaderfileModePair(sum1.c_str(),GL_COMPUTE_SHADER);
-  ostringstream s;
-  s << "LOCAL_SIZE_X " << localsize << "u" << endl;
+  ostringstream s,s2;
+  s << "LOCALSIZE " << gl::localSize << "u" << endl;
   shaderParams.push_back(s.str().c_str());
+  s2 << "BLOCKSIZE " << gl::blockSize << "u" << endl;
+  shaderParams.push_back(s2.str().c_str());
   GLuint rc=compileAndLinkShader(shaders,shaderParams,true,false,true);
   if(rc == 0) {
     GPUindexing=false; // Compute shaders are unavailable.
     if(settings::verbose > 2)
       cout << "No compute shader support" << endl;
   } else {
-    glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_COUNT,0,&maxgroups);
-    maxgroups=min(1024,maxgroups/(GLint) (localsize*localsize));
+//    glGetIntegeri_v(GL_MAX_COMPUTE_WORK_GROUP_COUNT,0,&maxgroups);
+//    maxgroups=min(1024,maxgroups/(GLint) (localSize*localSize));
     camp::sum1Shader=rc;
 
     shaders[0]=ShaderfileModePair(sum2.c_str(),GL_COMPUTE_SHADER);
-    camp::sum2Shader=compileAndLinkShader(shaders,shaderParams,true,false,true);
+    camp::sum2Shader=compileAndLinkShader(shaders,shaderParams,true,false,
+                                          true);
+
+    shaders[0]=ShaderfileModePair(sum2fast.c_str(),GL_COMPUTE_SHADER);
+    camp::sum2fastShader=compileAndLinkShader(shaders,shaderParams,true,false,
+                                              true);
 
     shaders[0]=ShaderfileModePair(sum3.c_str(),GL_COMPUTE_SHADER);
     camp::sum3Shader=compileAndLinkShader(shaders,shaderParams,true,false,
@@ -581,31 +592,44 @@ void initBlendShader()
   std::vector<ShaderfileModePair> shaders(2);
   std::vector<std::string> shaderParams;
 
-  ostringstream s,m2;
+  ostringstream s;
   s << "ARRAYSIZE " << maxSize << "u" << endl;
   shaderParams.push_back(s.str().c_str());
-  m2 << "m2 " << localsize;
-  shaderParams.push_back(m2.str().c_str());
   if(GPUindexing)
     shaderParams.push_back("GPUINDEXING");
+  if(GPUcompress)
+    shaderParams.push_back("GPUCOMPRESS");
   shaders[0]=ShaderfileModePair(screen.c_str(),GL_VERTEX_SHADER);
   shaders[1]=ShaderfileModePair(blend.c_str(),GL_FRAGMENT_SHADER);
   camp::blendShader=compileAndLinkShader(shaders,shaderParams,ssbo);
 }
 
+// Return the smallest power of 2 greater than or equal to n.
+inline GLuint ceilpow2(GLuint n)
+{
+  --n;
+  n |= n >> 1;
+  n |= n >> 2;
+  n |= n >> 4;
+  n |= n >> 8;
+  n |= n >> 16;
+  return ++n;
+}
+
 void initShaders()
 {
   Nlights=nlights == 0 ? 0 : max(Nlights,nlights);
   Nmaterials=max(Nmaterials,nmaterials);
 
   string zero=locateFile("shaders/zero.glsl");
+  string compress=locateFile("shaders/compress.glsl");
   string vertex=locateFile("shaders/vertex.glsl");
   string count=locateFile("shaders/count.glsl");
   string fragment=locateFile("shaders/fragment.glsl");
   string screen=locateFile("shaders/screen.glsl");
 
-  if(zero.empty() || vertex.empty() || fragment.empty() || screen.empty() ||
-     count.empty())
+  if(zero.empty() || compress.empty() || vertex.empty() || fragment.empty() ||
+     screen.empty() || count.empty())
     noShaders();
 
   if(GPUindexing)
@@ -624,6 +648,8 @@ void initShaders()
 #ifdef HAVE_SSBO
   if(GPUindexing)
     shaderParams.push_back("GPUINDEXING");
+  if(GPUcompress)
+    shaderParams.push_back("GPUCOMPRESS");
   shaders[1]=ShaderfileModePair(count.c_str(),GL_FRAGMENT_SHADER);
   camp::countShader=compileAndLinkShader(shaders,shaderParams,true);
   if(camp::countShader)
@@ -643,10 +669,6 @@ void initShaders()
     cout << "No SSBO support; order-independent transparency unavailable"
          << endl;
 
-  ostringstream m2;
-  m2 << "m2 " << localsize;
-  shaderParams.push_back(m2.str().c_str());
-
   shaders[1]=ShaderfileModePair(fragment.c_str(),GL_FRAGMENT_SHADER);
   shaderParams.push_back("MATERIAL");
   if(orthographic)
@@ -666,7 +688,7 @@ void initShaders()
   if(interlock) shaderParams.push_back("HAVE_INTERLOCK");
   camp::materialShader[0]=compileAndLinkShader(shaders,shaderParams,ssbo,
                                                interlock);
-  if(!camp::materialShader[0]) {
+  if(interlock && !camp::materialShader[0]) {
     shaderParams.pop_back();
     interlock=false;
     camp::materialShader[0]=compileAndLinkShader(shaders,shaderParams,ssbo);
@@ -686,7 +708,7 @@ void initShaders()
   shaderParams.pop_back();
 
   shaderParams.push_back("GENERAL");
-  if(Mode == 2)
+  if(Mode != 0)
     shaderParams.push_back("WIREFRAME");
   camp::generalShader[0]=compileAndLinkShader(shaders,shaderParams,ssbo,
                                               interlock);
@@ -700,8 +722,14 @@ void initShaders()
   shaderParams.clear();
 
   if(ssbo) {
-    if(!GPUindexing) {
-      shaders[0]=ShaderfileModePair(screen.c_str(),GL_VERTEX_SHADER);
+    if(GPUindexing)
+      shaderParams.push_back("GPUINDEXING");
+    shaders[0]=ShaderfileModePair(screen.c_str(),GL_VERTEX_SHADER);
+    shaders[1]=ShaderfileModePair(compress.c_str(),GL_FRAGMENT_SHADER);
+    camp::compressShader=compileAndLinkShader(shaders,shaderParams,ssbo);
+    if(GPUindexing)
+      shaderParams.pop_back();
+    else {
       shaders[1]=ShaderfileModePair(zero.c_str(),GL_FRAGMENT_SHADER);
       camp::zeroShader=compileAndLinkShader(shaders,shaderParams,ssbo);
     }
@@ -715,6 +743,7 @@ void deleteComputeShaders()
 {
   glDeleteProgram(camp::sum1Shader);
   glDeleteProgram(camp::sum2Shader);
+  glDeleteProgram(camp::sum2fastShader);
   glDeleteProgram(camp::sum3Shader);
 }
 
@@ -732,6 +761,7 @@ void deleteShaders()
     else
       glDeleteProgram(camp::zeroShader);
     glDeleteProgram(camp::countShader);
+    glDeleteProgram(camp::compressShader);
   }
 
   glDeleteProgram(camp::transparentShader);
@@ -743,6 +773,13 @@ void deleteShaders()
   glDeleteProgram(camp::pixelShader);
 }
 
+void resizeBlendShader(uint maxsize)
+{
+  gl::maxSize=ceilpow2(maxsize);
+  gl::deleteBlendShader();
+  gl::initBlendShader();
+}
+
 void setBuffers()
 {
   GLuint vao;
@@ -757,12 +794,15 @@ void setBuffers()
 
 #ifdef HAVE_SSBO
   glGenBuffers(1, &camp::offsetBuffer);
-  glGenBuffers(1, &camp::maxBuffer);
   if(GPUindexing) {
-    glGenBuffers(1, &camp::localSumBuffer);
     glGenBuffers(1, &camp::globalSumBuffer);
-  } else
-    glGenBuffers(1, &camp::countBuffer);
+    glGenBuffers(1, &camp::feedbackBuffer);
+  }
+  glGenBuffers(1, &camp::countBuffer);
+  if(GPUcompress) {
+    glGenBuffers(1, &camp::indexBuffer);
+    glGenBuffers(1, &camp::elementsBuffer);
+  }
   glGenBuffers(1, &camp::fragmentBuffer);
   glGenBuffers(1, &camp::depthBuffer);
   glGenBuffers(1, &camp::opaqueBuffer);
@@ -1121,15 +1161,8 @@ void togglefitscreen()
   fitscreen();
 }
 
-void initTimer()
-{
-  gettimeofday(&lasttime,NULL);
-  lastframetime=lasttime;
-}
-
 void idleFunc(void (*f)())
 {
-  initTimer();
   glutIdleFunc(f);
 }
 
@@ -1139,17 +1172,30 @@ void screen()
     fitscreen(false);
 }
 
-void nextframe(int)
+stopWatch frameTimer;
+
+void nextframe()
 {
 #ifdef HAVE_PTHREAD
   endwait(readySignal,readyLock);
 #endif
-  double framedelay=getSetting<double>("framedelay");
-  if(framedelay > 0)
-    usleep((unsigned int) (1000.0*framedelay+0.5));
+  double delay=getSetting<double>("framerate");
+  if(delay != 0.0) delay=1.0/delay;
+  double seconds=frameTimer.seconds(true);
+  delay -= seconds;
+  if(delay > 0) {
+    timespec req;
+    timespec rem;
+    req.tv_sec=(unsigned int) delay;
+    req.tv_nsec=(unsigned int) (1.0e9*(delay-req.tv_sec));
+    while(nanosleep(&req,&rem) < 0 && errno == EINTR)
+      req=rem;
+  }
   if(Step) Animate=false;
 }
 
+stopWatch Timer;
+
 void display()
 {
   if(queueScreen) {
@@ -1161,9 +1207,9 @@ void display()
   drawscene(Width,Height);
   if(fps) {
     if(framecount < 20) // Measure steady-state framerate
-      seconds();
+      Timer.reset();
     else {
-      double s=seconds();
+      double s=Timer.seconds(true);
       if(s > 0.0) {
         double rate=1.0/s;
         S.add(rate);
@@ -1179,18 +1225,7 @@ void display()
 #ifdef HAVE_PTHREAD
   if(glthread && Animate) {
     queueExport=false;
-    double delay=1.0/getSetting<double>("framerate");
-    timeval tv;
-    gettimeofday(&tv,NULL);
-    double seconds=tv.tv_sec-lastframetime.tv_sec+
-      ((double) tv.tv_usec-lastframetime.tv_usec)/1000000.0;
-    lastframetime=tv;
-    double milliseconds=1000.0*(delay-seconds);
-    double framedelay=getSetting<double>("framedelay");
-    if(framedelay > 0) milliseconds -= framedelay;
-    if(milliseconds > 0)
-      glutTimerFunc((int) (milliseconds+0.5),nextframe,0);
-    else nextframe(0);
+    nextframe();
   }
 #endif
   if(queueExport) {
@@ -1208,7 +1243,6 @@ void display()
 void update()
 {
   glutDisplayFunc(display);
-  Animate=getSetting<bool>("autoplay");
   glutShowWindow();
   if(Zoom != lastzoom) remesh=true;
   lastzoom=Zoom;
@@ -1289,16 +1323,14 @@ void shift(int x, int y)
 
 void pan(int x, int y)
 {
-  if(orthographic) {
-    double Zoominv=1.0/Zoom;
-    X += (x-x0)*Zoominv;
-    Y += (y0-y)*Zoominv;
-  } else {
+  if(orthographic)
+    shift(x,y);
+  else {
     cx += (x-x0)*(xmax-xmin)/Width;
     cy += (y0-y)*(ymax-ymin)/Height;
+    x0=x; y0=y;
+    update();
   }
-  x0=x; y0=y;
-  update();
 }
 
 void zoom(int x, int y)
@@ -1858,6 +1890,12 @@ void init_osmesa()
 
 #endif /* HAVE_GL */
 
+bool NVIDIA()
+{
+  char *GLSL_VERSION=(char *) glGetString(GL_SHADING_LANGUAGE_VERSION);
+  return string(GLSL_VERSION).find("NVIDIA") != string::npos;
+}
+
 // angle=0 means orthographic.
 void glrender(const string& prefix, const picture *pic, const string& format,
               double width, double height, double angle, double zoom,
@@ -1866,14 +1904,9 @@ void glrender(const string& prefix, const picture *pic, const string& format,
               double *background, size_t nlightsin, triple *lights,
               double *diffuse, double *specular, bool view, int oldpid)
 {
+  gettimeofday(&lasttime,NULL);
   Iconify=getSetting<bool>("iconify");
 
-#if defined(HAVE_COMPUTE_SHADER) && !defined(HAVE_LIBOSMESA)
-  GPUindexing=getSetting<bool>("GPUindexing");
-#else
-  GPUindexing=false;
-#endif
-
   if(zoom == 0.0) zoom=1.0;
 
   Prefix=prefix;
@@ -2120,11 +2153,22 @@ void glrender(const string& prefix, const picture *pic, const string& format,
 
   initialized=true;
 
+#if defined(HAVE_COMPUTE_SHADER) && !defined(HAVE_LIBOSMESA)
+  GPUindexing=getSetting<bool>("GPUindexing");
+  GPUcompress=getSetting<bool>("GPUcompress");
+#else
+  GPUindexing=false;
+  GPUcompress=false;
+#endif
+
   GLint val;
   glGetIntegerv(GL_MAX_UNIFORM_BLOCK_SIZE,&val);
 
-  if(GPUindexing)
-    localsize=getSetting<Int>("GPUlocalSize");
+  if(GPUindexing) {
+    gl::localSize=getSetting<Int>("GPUlocalSize");
+    gl::blockSize=getSetting<Int>("GPUblockSize");
+    gl::groupSize=gl::localSize*gl::blockSize;
+  }
 
   Maxmaterials=val/sizeof(Material);
   if(nmaterials > Maxmaterials) nmaterials=Maxmaterials;
@@ -2275,170 +2319,54 @@ void clearCount()
   glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
 }
 
-void initPartialSums()
+void compressCount()
 {
-  gl::gs=gl::localsize*gl::g;
-  gl::gs2=gl::localsize*gl::gs;
-  gl::processors=gl::localsize*gl::gs2;
-  GLuint zero=0;
-  glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::localSumBuffer);
-  glBufferData(GL_SHADER_STORAGE_BUFFER,(gl::processors+gl::gs2+2)*sizeof(GLuint),NULL,
-               GL_DYNAMIC_DRAW);
-  glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                    GL_UNSIGNED_BYTE,&zero);
-  glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,camp::localSumBuffer);
-
-  glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::globalSumBuffer);
-  glBufferData(GL_SHADER_STORAGE_BUFFER,(1+gl::gs)*sizeof(GLuint),NULL,
-               GL_DYNAMIC_DRAW);
-  glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                    GL_UNSIGNED_BYTE,&zero);
-  glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,camp::globalSumBuffer);
+  glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
+  glUseProgram(compressShader);
+  gl::lastshader=compressShader;
+  glUniform1ui(glGetUniformLocation(compressShader,"width"),gl::Width);
+  fpu_trap(false); // Work around FE_INVALID
+  glDrawArrays(GL_TRIANGLES, 0, 3);
+  fpu_trap(settings::trap());
+  glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
 }
 
-GLuint partialSums()
+void partialSums(bool readSize=false)
 {
-  GLuint fragments;
-  // Compute local partial sums on the GPU
+  // Compute partial sums on the GPU
   glUseProgram(sum1Shader);
-  glUniform1ui(glGetUniformLocation(sum1Shader,"elements"),gl::pixels);
-  glDispatchCompute(gl::gs2,1,1);
-
-  glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
-
-  glUseProgram(sum2Shader);
-  GLuint offset2=gl::processors+1;
-  glUniform1ui(glGetUniformLocation(sum2Shader,"offset2"),offset2);
-  glDispatchCompute(gl::gs,1,1);
-
-  glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
-
-  glUseProgram(sum3Shader);
-  glUniform1ui(glGetUniformLocation(sum3Shader,"offset2"),offset2);
   glDispatchCompute(gl::g,1,1);
 
-  glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
-  // Compute global partial sums, including number of fragments, on the CPU
-  GLuint *sum=(GLuint *) (glMapBuffer(GL_SHADER_STORAGE_BUFFER,
-                                      GL_READ_WRITE));
-  fragments=0;
-  for(GLint i=1; i <= gl::gs; ++i) {
-    fragments += sum[i];
-    sum[i]=fragments;
+  if(gl::elements <= gl::groupSize*gl::groupSize)
+    glUseProgram(sum2fastShader);
+  else {
+    glUseProgram(sum2Shader);
+    glUniform1ui(glGetUniformLocation(sum2Shader,"blockSize"),
+                 gl::ceilquotient(gl::g,gl::localSize));
   }
-  glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
-  return fragments;
+  glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
+  glDispatchCompute(1,1,1);
+
+  glUseProgram(sum3Shader);
+  glUniform1ui(glGetUniformLocation(sum3Shader,"final"),gl::elements-1);
+  glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
+  glDispatchCompute(gl::g,1,1);
 }
 
-void refreshBuffers()
+void resizeFragmentBuffer()
 {
-  GLuint zero=0;
-  GLuint fragments;
-  gl::pixels=gl::Width*gl::Height;
+  if(GPUindexing) {
+    glMemoryBarrier(GL_BUFFER_UPDATE_BARRIER_BIT);
 
-  if(initSSBO) {
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,
-                 (GPUindexing ? 2:1)*gl::pixels*sizeof(GLuint),
-                 NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,camp::offsetBuffer);
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                      GL_UNSIGNED_BYTE,&zero);
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+    GLuint *feedback=(GLuint *) glMapBuffer(GL_SHADER_STORAGE_BUFFER,GL_READ_ONLY);
 
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::maxBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,sizeof(GLuint),
-                 NULL,GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,camp::maxBuffer);
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                      GL_UNSIGNED_BYTE,&zero);
+    GLuint maxDepth=feedback[0];
+    if(maxDepth > gl::maxSize)
+      gl::resizeBlendShader(maxDepth);
 
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::opaqueBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(glm::vec4),NULL,
-                 GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,6,camp::opaqueBuffer);
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::opaqueDepthBuffer);
-    glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(GLfloat),NULL,
-                 GL_DYNAMIC_DRAW);
-    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,7,camp::opaqueDepthBuffer);
-    const GLfloat zerof=0.0;
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32F,GL_RED,GL_FLOAT,&zerof);
-
-    if(GPUindexing) {
-      double Tmin=HUGE_VAL;
-      GLuint G=1;
-      for(gl::g=2; gl::g <= gl::maxgroups;
-          gl::g *= 2) {
-        initPartialSums();
-        partialSums();
-        seconds();
-        partialSums();
-        double T=seconds();
-        if(T < Tmin) {
-          Tmin=T;
-          G=gl::g;
-        }
-      }
-      gl::g=G;
-      initPartialSums();
-    } else {
-      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
-      glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(GLuint),
-                   NULL,GL_DYNAMIC_DRAW);
-      glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,camp::countBuffer);
-      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                        GL_UNSIGNED_BYTE,&zero);
-    }
-    initSSBO=false;
-  }
-
-  if(GPUindexing && gl::exporting) {
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
-    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
-                      GL_UNSIGNED_BYTE,&zero);
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::globalSumBuffer);
-  }
-
-  // Determine the fragment offsets
-
-  if(!interlock) {
-    drawBuffer(material1Data,countShader);
-    drawBuffer(materialData,countShader);
-    drawBuffer(colorData,countShader,true);
-    drawBuffer(triangleData,countShader,true);
-  }
-
-  glDepthMask(GL_FALSE); // Don't write to depth buffer
-  glDisable(GL_MULTISAMPLE);
-  drawBuffer(transparentData,countShader,true);
-  glEnable(GL_MULTISAMPLE);
-  glDepthMask(GL_TRUE); // Write to depth buffer
-
-  if(GPUindexing)
-    fragments=partialSums();
-  else {
-    // Compute partial sums on the CPU
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
-    GLuint *count=((GLuint *) glMapBuffer(GL_SHADER_STORAGE_BUFFER,
-                                          GL_READ_ONLY));
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
-    GLuint *offset=(GLuint *) glMapBuffer(GL_SHADER_STORAGE_BUFFER,
-                                          GL_WRITE_ONLY);
-
-    size_t Offset=offset[0]=count[0];
-    for(size_t i=1; i < gl::pixels; ++i)
-      offset[i]=Offset += count[i];
-    fragments=Offset;
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
+    fragments=feedback[1];
     glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
-
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
-    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
-
-    clearCount();
   }
 
   if(fragments > maxFragments) {
@@ -2455,10 +2383,170 @@ void refreshBuffers()
                  NULL,GL_DYNAMIC_DRAW);
     glBindBufferBase(GL_SHADER_STORAGE_BUFFER,5,camp::depthBuffer);
 
-    if(GPUindexing)
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+  }
+}
+
+void refreshBuffers()
+{
+  GLuint zero=0;
+  gl::pixels=(gl::Width+1)*(gl::Height+1);
+
+  if(initSSBO) {
+    gl::processors=1;
+
+    GLuint Pixels;
+    if(GPUindexing) {
+      GLuint G=gl::ceilquotient(gl::pixels,gl::groupSize);
+      Pixels=gl::groupSize*G;
+
+      GLuint globalSize=gl::localSize*gl::ceilquotient(G,gl::localSize);
       glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::globalSumBuffer);
+      glBufferData(GL_SHADER_STORAGE_BUFFER,globalSize*sizeof(GLuint),NULL,
+                   GL_DYNAMIC_READ);
+      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32UI,GL_RED_INTEGER,
+                        GL_UNSIGNED_INT,&zero);
+      glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,camp::globalSumBuffer);
+    } else Pixels=gl::pixels;
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,(Pixels+2)*sizeof(GLuint),
+                 NULL,GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,camp::offsetBuffer);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,(Pixels+2)*sizeof(GLuint),
+                 NULL,GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,camp::countBuffer);
+
+    if(GPUcompress) {
+      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::indexBuffer);
+      glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(GLuint),
+                   NULL,GL_DYNAMIC_DRAW);
+      glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,camp::indexBuffer);
+
+      GLuint one=1;
+      glBindBuffer(GL_ATOMIC_COUNTER_BUFFER,camp::elementsBuffer);
+      glBufferData(GL_ATOMIC_COUNTER_BUFFER,sizeof(GLuint),&one,
+                   GL_DYNAMIC_DRAW);
+      glBindBufferBase(GL_ATOMIC_COUNTER_BUFFER,0,camp::elementsBuffer);
+    } else
+      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
+                        GL_UNSIGNED_BYTE,&zero); // Clear the count buffer
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::opaqueBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,gl::pixels*sizeof(glm::vec4),NULL,
+                 GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,6,camp::opaqueBuffer);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::opaqueDepthBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,
+                 sizeof(GLuint)+gl::pixels*sizeof(GLfloat),NULL,
+                 GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,7,camp::opaqueDepthBuffer);
+    const GLfloat zerof=0.0;
+    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32F,GL_RED,GL_FLOAT,&zerof);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+    glBufferData(GL_SHADER_STORAGE_BUFFER,2*sizeof(GLuint),NULL,
+                 GL_DYNAMIC_DRAW);
+    glBindBufferBase(GL_SHADER_STORAGE_BUFFER,8,camp::feedbackBuffer);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+    initSSBO=false;
   }
 
+  // Determine the fragment offsets
+
+  if(gl::exporting && GPUindexing && !GPUcompress) {
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
+    glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R8UI,GL_RED_INTEGER,
+                      GL_UNSIGNED_BYTE,&zero);
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::feedbackBuffer);
+  }
+
+  if(!interlock) {
+    drawBuffer(material1Data,countShader);
+    drawBuffer(materialData,countShader);
+    drawBuffer(colorData,countShader,true);
+    drawBuffer(triangleData,countShader,true);
+  }
+
+  glDepthMask(GL_FALSE); // Don't write to depth buffer
+  glDisable(GL_MULTISAMPLE);
+  drawBuffer(transparentData,countShader,true);
+  glEnable(GL_MULTISAMPLE);
+  glDepthMask(GL_TRUE); // Write to depth buffer
+
+  if(GPUcompress) {
+    compressCount();
+    GLuint *p=(GLuint *) glMapBuffer(GL_ATOMIC_COUNTER_BUFFER,GL_READ_WRITE);
+    gl::elements=GPUindexing ? p[0] : p[0]-1;
+    p[0]=1;
+    glUnmapBuffer(GL_ATOMIC_COUNTER_BUFFER);
+    if(gl::elements == 0) return;
+  } else
+    gl::elements=gl::pixels;
+
+  if(GPUindexing) {
+    gl::g=gl::ceilquotient(gl::elements,gl::groupSize);
+    gl::elements=gl::groupSize*gl::g;
+
+    if(settings::verbose > 3) {
+      static bool first=true;
+      if(first) {
+        partialSums();
+        first=false;
+      }
+      unsigned int N=10000;
+      stopWatch Timer;
+      for(unsigned int i=0; i < N; ++i)
+        partialSums();
+      glFinish();
+      double T=Timer.seconds()/N;
+      cout << "elements=" << gl::elements << endl;
+      cout << "Tmin (ms)=" << T*1e3 << endl;
+      cout << "Megapixels/second=" << gl::elements/T/1e6 << endl;
+    }
+
+    partialSums(true);
+  } else {
+    size_t size=gl::elements*sizeof(GLuint);
+
+    // Compute partial sums on the CPU
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
+    GLuint *p=(GLuint *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
+                                          0,size+sizeof(GLuint),
+                                              GL_MAP_READ_BIT);
+    GLuint maxsize=p[0];
+    GLuint *count=p+1;
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
+    GLuint *offset=(GLuint *) glMapBufferRange(GL_SHADER_STORAGE_BUFFER,
+                                               sizeof(GLuint),size,
+                                               GL_MAP_WRITE_BIT);
+
+    size_t Offset=offset[0]=count[0];
+    for(size_t i=1; i < gl::elements; ++i)
+      offset[i]=Offset += count[i];
+    fragments=Offset;
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::offsetBuffer);
+    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
+
+    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
+    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
+
+    if(gl::exporting) {
+      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::countBuffer);
+      glClearBufferData(GL_SHADER_STORAGE_BUFFER,GL_R32UI,GL_RED_INTEGER,
+                        GL_UNSIGNED_INT,&zero);
+    } else
+      clearCount();
+
+    if(maxsize > gl::maxSize)
+      gl::resizeBlendShader(maxsize);
+  }
   gl::lastshader=-1;
 }
 
@@ -2471,17 +2559,6 @@ void setUniforms(vertexBuffer& data, GLint shader)
 
     if(normal)
       glUniform1ui(glGetUniformLocation(shader,"width"),gl::Width);
-
-    if(camp::ssbo && GPUindexing &&
-       (shader == transparentShader || !interlock)) {
-      glUniform1ui(glGetUniformLocation(shader,"pixels"),gl::pixels);
-      GLuint offset2=1+gl::processors;
-      glUniform1ui(glGetUniformLocation(shader,"offset2"),offset2);
-      GLuint m=GPUindexing ? gl::pixels/gl::processors : 0;
-      GLuint r=gl::pixels-m*gl::processors;
-      glUniform1ui(glGetUniformLocation(shader,"m1"),m);
-      glUniform1ui(glGetUniformLocation(shader,"r"),r);
-    }
   }
 
   glUniformMatrix4fv(glGetUniformLocation(shader,"projViewMat"),1,GL_FALSE,
@@ -2628,18 +2705,6 @@ void drawTriangle()
   triangleData.clear();
 }
 
-// Return the smallest power of 2 greater than or equal to n.
-inline GLuint ceilpow2(GLuint n)
-{
-  --n;
-  n |= n >> 1;
-  n |= n >> 2;
-  n |= n >> 4;
-  n |= n >> 8;
-  n |= n >> 16;
-  return ++n;
-}
-
 void aBufferTransparency()
 {
   // Collect transparent fragments
@@ -2652,15 +2717,6 @@ void aBufferTransparency()
   glUseProgram(blendShader);
   gl::lastshader=blendShader;
   glUniform1ui(glGetUniformLocation(blendShader,"width"),gl::Width);
-  if(GPUindexing) {
-    glUniform1ui(glGetUniformLocation(blendShader,"pixels"),gl::pixels);
-    GLuint offset2=gl::processors+1;
-    glUniform1ui(glGetUniformLocation(blendShader,"offset2"),offset2);
-    GLuint m=GPUindexing ? gl::pixels/gl::processors : 0;
-    GLuint r=gl::pixels-m*gl::processors;
-    glUniform1ui(glGetUniformLocation(blendShader,"m1"),m);
-    glUniform1ui(glGetUniformLocation(blendShader,"r"),r);
-  }
   glUniform4f(glGetUniformLocation(blendShader,"background"),
               gl::Background[0],gl::Background[1],gl::Background[2],
               gl::Background[3]);
@@ -2669,22 +2725,6 @@ void aBufferTransparency()
   glDrawArrays(GL_TRIANGLES,0,3);
   fpu_trap(settings::trap());
   transparentData.clear();
-
-  if(gl::maxSize > 0) {
-    glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::maxBuffer);
-    GLuint *p=(GLuint *) glMapBuffer(GL_SHADER_STORAGE_BUFFER,GL_READ_WRITE);
-    gl::maxSize=p[0];
-    p[0]=0;
-    glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);
-    if(gl::maxSize > 0) {
-      gl::maxSize=ceilpow2(gl::maxSize);
-      gl::deleteBlendShader();
-      gl::initBlendShader();
-    }
-    if(GPUindexing)
-      glBindBuffer(GL_SHADER_STORAGE_BUFFER,camp::globalSumBuffer);
-  }
-
   glEnable(GL_DEPTH_TEST);
 }
 
@@ -2712,7 +2752,10 @@ void drawBuffers()
   if(camp::ssbo) {
     if(transparent) {
       refreshBuffers();
-      if(!interlock) gl::copied=true;
+      if(!interlock) {
+        resizeFragmentBuffer();
+        gl::copied=true;
+      }
     }
   }
 
@@ -2724,6 +2767,7 @@ void drawBuffers()
 
   if(transparent) {
     gl::copied=true;
+    if(interlock) resizeFragmentBuffer();
     drawTransparent();
   }
   Opaque=0;
